import inspect
import subprocess
from pathlib import Path
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK, SLK_HELPERS
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import run_slk, which

__all__ = [
    "json2hsm_raw",
    "hsm2json_raw",
    "metadata_raw",
    "tag_raw",
]


def json2hsm_raw(
    json_file: Union[str, Path, None] = None,
    restart_file: Union[str, Path, None] = None,
    schema: Union[str, list, None] = None,
    expect_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    ignore_non_existing_metadata_fields: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_update: bool = False,
    use_res_id: bool = False,
    skip_bad_metadata_sets: bool = False,
    print_summary: bool = False,
    print_json_summary: bool = False,
    json_string: Union[str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """
    Read metadata from JSON file and write them to archived files into HSM. Use absolute paths from metadata records
    to identify target files.

    :param json_file: JSON input file containing metadata; incompatible with json_string [default: None]
    :type json_file: str or Path-like or None
    :param restart_file: set a restart file in which the processed metadata entries are listed [default: None]
    :type restart_file: str or Path-like or None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: str or list
    :param expect_json_lines: read JSON-lines from file instead of JSON
    :type expect_json_lines: bool
    :param verbose: verbose mode
    :type verbose: bool
    :param quiet: quiet mode
    :type quiet: bool
    :param ignore_non_existing_metadata_fields: throw no error if metadata field does not exist in StrongLink
    :type ignore_non_existing_metadata_fields: bool
    :param write_mode: select write mode for metadata: OVERWRITE, KEEP, ERROR, CLEAN
    :type write_mode: str
    :param use_res_id: use resource_id instead of path to identify file
    :type use_res_id: bool
    :param skip_bad_metadata_sets: skip damaged / incomplete metadata sets [default: throw error]
    :type skip_bad_metadata_sets: bool
    :param instant_metadata_record_update: False (default): read metadata records of all files and import into
                StrongLink afterwards; True: import each metadata record after it has been read
    :type instant_metadata_record_update: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :param print_summary: print a summary
    :type print_summary: bool
    :param print_json_summary: print a summary as JSON
    :type print_json_summary: bool
    :param json_string: provide a json string instead of a json file; incompatible with json_file
    :type json_string: str
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "json2hsm"]
    if json_file is not None:
        if not isinstance(json_file, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'json_file' needs to be 'str' or Path-like but got "
                + f"'{type(json_file).__name__}'."
            )
        slk_call.append(str(json_file))
    if restart_file is not None:
        if not isinstance(restart_file, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'restart_file' needs to be 'str' or Path-like but got "
                + f"'{type(restart_file).__name__}'."
            )
        if " " in str(restart_file):
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: 'restart_file' is not allowed to contain spaces"
            )
        slk_call.append("--restart-file")
        slk_call.append(str(restart_file))
    if expect_json_lines:
        slk_call.append("--expect-json-lines")
    if ignore_non_existing_metadata_fields:
        slk_call.append("--ignore-non-existing-metadata-fields")
    if verbose:
        slk_call.append("--verbose")
    if quiet:
        slk_call.append("-q")
    if use_res_id:
        slk_call.append("--use-res-id")
    if skip_bad_metadata_sets:
        slk_call.append("-k")
    if schema is not None:
        slk_call.append("--schema")
        if isinstance(schema, str):
            slk_call.append(schema)
        elif isinstance(schema, list):
            slk_call.append(",".join(schema))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'schema' has wrong type; need 'str' or 'list' of 'str' but got "
                + f"{type(schema).__name__}"
            )
    if write_mode is not None:
        slk_call.append("--write-mode")
        slk_call.append(write_mode)
    if instant_metadata_record_update:
        slk_call.append("--instant-metadata-record-update")
    if print_summary:
        slk_call.append("--print-summary")
    if print_json_summary:
        slk_call.append("--print-json-summary")
    if json_string is not None:
        slk_call.append("--json-string")
        slk_call.append("'" + json_string + "'")

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def hsm2json_raw(
    resources: Union[str, list, Path, None] = None,
    search_id: int = -1,
    recursive: bool = False,
    outfile: Union[str, Path, None] = None,
    restart_file: Union[str, Path, None] = None,
    schema: Union[str, list, None] = None,
    write_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    write_mode: Union[str, None] = None,
    print_summary: bool = False,
    print_json_summary: bool = False,
    write_compact_json: bool = False,
    instant_metadata_record_output: bool = False,
    print_hidden: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """
    Extract metadata from HSM file(s) and return them in JSON structure

    :param resources: list of resources to be searched for [default: None]
    :type resources: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: int
    :param recursive: export metadata from all files in gns_path recursively
    :type recursive: bool
    :param outfile: Write the output into a file instead to the stdout [default: None]
    :type outfile: str or Path-like or None
    :param restart_file: set a restart file in which the processed metadata entries are listed [default: None]
    :type restart_file: str or Path-like or None
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: str, list or None
    :param write_json_lines: write JSON-lines instead of JSON
    :type write_json_lines: bool = False
    :param verbose: verbose mode
    :type verbose: bool = False
    :param quiet: quiet mode
    :type quiet: bool = False
    :param write_mode: applies when 'output' is set; possible values: OVERWRITE, ERROR
    :type write_mode: str = None
    :param print_summary: print summary on how many metadata records have been processed [default: False]
    :type print_summary: bool = False
    :param print_json_summary: print summary on how many metadata records have been processed as JSON [default: False]
    :type print_json_summary: bool = False
    :param write_compact_json: write JSON metadata in a compact form with less line breaks [default: False]
    :type write_compact_json: bool = False
    :param instant_metadata_record_output: False (default): read metadata of all files and write/print out afterwards;
                True: write/print each metadata record after it has been read (requires 'write_json_lines')
    :type instant_metadata_record_output: bool
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "hsm2json"]
    if resources is not None:
        if isinstance(resources, Path):
            slk_call.append(str(resources))
        elif isinstance(resources, str):
            if resources != "":
                slk_call.extend(resources.split(" "))
        elif isinstance(resources, list):
            if len(resources) > 0:
                if not all([isinstance(r, (str, Path)) for r in resources]):
                    raise TypeError(
                        f"pyslk.{inspect.stack()[0][3]}: if argument 'resources' is of type 'list' its items need to be"
                        + "of type 'str' or Path-like but got type(s): "
                        + f"'{', '.join([type(r).__name__ for r in resources if not isinstance(r, (str, Path))])}'."
                    )
                slk_call.extend(resources)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'resources' needs to be 'str', 'list' (of 'str' or "
                + f"Path-like) or Path-like but got '{type(resources).__name__}'."
            )
    if search_id != -1:
        slk_call.append("--search-id")
        slk_call.append(str(search_id))
    if recursive:
        slk_call.append("-R")
    if outfile is not None:
        if not isinstance(outfile, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'outfile' needs to be 'str' or Path-like but got "
                + f"'{type(outfile).__name__}'."
            )
        if " " in str(outfile):
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: 'outfile' is not allowed to contain spaces"
            )
        slk_call.append("--outfile")
        slk_call.append(str(outfile))
    if restart_file is not None:
        if not isinstance(restart_file, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'restart_file' needs to be 'str' or Path-like but got "
                + f"'{type(restart_file).__name__}'."
            )
        if " " in str(restart_file):
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: 'restart_file' is not allowed to contain spaces"
            )
        slk_call.append("--restart-file")
        slk_call.append(str(restart_file))
    if write_json_lines:
        slk_call.append("--write-json-lines")
    if verbose:
        slk_call.append("--verbose")
    if quiet:
        slk_call.append("-q")
    if instant_metadata_record_output:
        slk_call.append("--instant-metadata-record-output")
    if print_summary:
        slk_call.append("--print-summary")
    if print_json_summary:
        slk_call.append("--print-json-summary")
    if write_compact_json:
        slk_call.append("--write-compact-json")
    if print_hidden:
        slk_call.append("--print-hidden")
    if schema is not None:
        slk_call.append("--schema")
        if isinstance(schema, str):
            slk_call.append(schema)
        elif isinstance(schema, list):
            slk_call.append(",".join(schema))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'schema' has wrong type; need 'str' or 'list' of 'str' but got "
                + f"{type(schema).__name__}"
            )
    if write_mode is not None:
        slk_call.append("--write-mode")
        slk_call.append(write_mode)

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def metadata_raw(
    resource_path: Union[str, Path],
    alternative_output_format: bool = False,
    print_hidden: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get metadata

    :param resource_path: resource (full path)
    :type resource_path: str or Path
    :param alternative_output_format: print the name of the metadata schema in front of each field name [default: False]
    :type alternative_output_format: bool
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" but is '
            + f"{type(resource_path).__name__}."
        )

    slk_call = [SLK_HELPERS, "metadata", str(resource_path)]

    if alternative_output_format:
        slk_call.append("--alternative-output-format")
    if print_hidden:
        slk_call.append("--print-hidden")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tag_raw(
    path_or_id: Union[str, int, Path],
    metadata: Union[dict, None] = None,
    recursive: bool = False,
    display: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Apply metadata to the namespace and child resources

    :param path_or_id: search id or gns path of resources to retrieve
    :type path_or_id: str or int or Path
    :param metadata: dict that holds as keys "[metadata schema].[field]" and
                     as values the metadata values
    :type metadata: dict
    :param recursive: use the -R flag to tag recursively, Default: False
    :type recursive: bool
    :param display: print metadata as JSON (ignore "metadata" if this is True), Default: False
    :type display: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    if metadata is None and not display:
        raise ValueError(
            f"slk.{inspect.stack()[0][3]}: neither 'metadata' is set nor 'display' is 'true'; however, either argument "
            + "'metadata' has to be set (== set metadata) or 'display' has to be 'true' (== print metadata)"
        )
    if metadata is not None and display:
        raise ValueError(
            f"slk.{inspect.stack()[0][3]}: 'metadata' is set and 'display' is 'true'; however, argument 'metadata' has "
            + "to be set (== set metadata) XOR 'display' has to be 'true' (== print metadata); not both"
        )

    slk_call = [SLK, "tag"]
    if recursive:
        slk_call.append("-R")
    if isinstance(path_or_id, (str, Path, int)):
        slk_call.append(str(path_or_id))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' needs to be 'str', 'int' or path-like but got "
            + f"'{type(path_or_id).__name__}'"
        )
    if display:
        slk_call.append("-display")
    else:
        if isinstance(metadata, dict):
            if len(metadata) == 0:
                raise ValueError(
                    f"slk.{inspect.stack()[0][3]}: argument 'metadata' is an empty dictionary; needs to have at "
                    + "least one key-value pair"
                )
            for k, v in metadata.items():
                slk_call.append(k + "=" + v)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'metadata' needs to be 'dict' but got "
                + f"'{type(metadata).__name__}'"
            )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



from .gen_queries import gen_file_query, gen_file_query_as_dict, gen_search_query
from .gfbt import (
    _check_input_gfbt,
    _gen_group_collection,
    _group_retrieve,
    count_tapes,
    count_tapes_with_multi_tape_files,
    count_tapes_with_single_tape_files,
    group_files_by_tape,
)
from .metadata import (
    get_metadata,
    get_tag,
    hsm2json,
    hsm2json_dict,
    hsm2json_file,
    json2hsm,
    json_dict2hsm,
    json_file2hsm,
    json_str2hsm,
    set_tag,
)
from .resource import (
    access_hsm,
    arch_size,
    chgrp,
    chmod,
    chown,
    delete,
    get_checksum,
    get_resource_id,
    get_resource_mtime,
    get_resource_permissions,
    get_resource_size,
    get_resource_tape,
    has_no_flag_partial,
    has_no_flag_partial_details,
    makedirs,
    mkdir,
    move,
    rename,
)
from .resource_extras import get_resource_type, is_file, is_namespace, resource_exists
from .storage import (
    _cached,
    _on_tape,
    get_rcrs,
    get_storage_information,
    get_tape_barcode,
    get_tape_id,
    get_tape_status,
    is_cached,
    is_cached_details,
    is_on_tape,
    is_on_tape_details,
    is_tape_available,
    tape_exists,
    tape_status,
)

__all__ = [
    "_cached",
    "_check_input_gfbt",
    "_gen_group_collection",
    "_group_retrieve",
    "_on_tape",
    "access_hsm",
    "arch_size",
    "chgrp",
    "chmod",
    "chown",
    "count_tapes",
    "count_tapes_with_multi_tape_files",
    "count_tapes_with_single_tape_files",
    "delete",
    "gen_file_query",
    "gen_file_query_as_dict",
    "gen_search_query",
    "get_checksum",
    "get_metadata",
    "get_rcrs",
    "get_resource_id",
    "get_resource_mtime",
    "get_resource_permissions",
    "get_resource_size",
    "get_resource_tape",
    "get_resource_type",
    "get_storage_information",
    "get_tag",
    "get_tape_barcode",
    "get_tape_id",
    "get_tape_status",
    "group_files_by_tape",
    "has_no_flag_partial",
    "has_no_flag_partial_details",
    "hsm2json",
    "hsm2json_dict",
    "hsm2json_file",
    "is_cached",
    "is_cached_details",
    "is_file",
    "is_namespace",
    "is_on_tape",
    "is_on_tape_details",
    "is_tape_available",
    "json2hsm",
    "json_dict2hsm",
    "json_file2hsm",
    "json_str2hsm",
    "makedirs",
    "mkdir",
    "move",
    "rename",
    "resource_exists",
    "set_tag",
    "tape_exists",
    "tape_status",
]



import inspect
import subprocess
from pathlib import Path
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK_HELPERS
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import run_slk, which

__all__ = [
    "tape_status_raw",
    "tape_id_raw",
    "tape_barcode_raw",
    "tape_exists_raw",
    "tape_library_raw",
    "iscached_raw",
    "is_on_tape_raw",
    "group_files_by_tape_raw",
    "print_rcrs_raw",
]


def tape_status_raw(
    tape_id: Union[int, str, None] = None,
    tape_barcode: Union[str, None] = None,
    details: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check the status of a tape

    :param tape_id: id of a tape in the tape library
    :type tape_id: int or str or None
    :param tape_barcode: barcode of a tape in the tape library
    :type tape_barcode: str or None
    :param details: print a more detailed description of the retrieval status
    :type details: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "tape_status"]

    if tape_id is not None:
        if not isinstance(tape_id, (int, str)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'tape_id' needs to be 'int' or 'str' but got "
                + f"'{type(tape_id).__name__}'."
            )
        slk_call.append(str(tape_id))
    if tape_barcode is not None:
        if not isinstance(tape_barcode, str):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'tape_barcode' needs to be 'str' but got "
                + f"'{type(tape_barcode).__name__}'."
            )
        slk_call.append("--tape-barcode")
        slk_call.append(tape_barcode)
    if details:
        slk_call.append("--details")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tape_id_raw(
    tape_barcode: str, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """return tape id for provided tape barcode

    :param tape_barcode: barcode of a tape in the tape library
    :type tape_barcode: str
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(tape_barcode, str):
        raise TypeError(
            f'slk.{inspect.stack()[0][3]}: argument "tape_id" needs to be "str" but got '
            + f'"{type(tape_barcode).__name__}".'
        )

    slk_call = [SLK_HELPERS, "tape_id", tape_barcode]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tape_barcode_raw(
    tape_id: Union[int, str], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """return tape barcode for provided tape id

    :param tape_id: id of a tape in the tape library
    :type tape_id: int or str
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(tape_id, (int, str)):
        raise TypeError(
            f'slk.{inspect.stack()[0][3]}: argument "tape_id" needs to be "int" or "str" but got '
            + f'"{type(tape_id).__name__}".'
        )

    slk_call = [SLK_HELPERS, "tape_barcode", str(tape_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tape_exists_raw(
    tape_id: Union[int, str, None] = None,
    tape_barcode: Union[str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if tape exists

    :param tape_id: id of a tape in the tape library
    :type tape_id: int or str or None
    :param tape_barcode: barcode of a tape in the tape library
    :type tape_barcode: str or None
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "tape_exists"]

    if tape_id is not None:
        if not isinstance(tape_id, (int, str)):
            raise TypeError(
                f"slk.{inspect.stack()[0][3]}: argument 'tape_id' needs to be 'int' or 'str' but got "
                + f"'{type(tape_id).__name__}'."
            )
        slk_call.append(str(tape_id))
    if tape_barcode is not None:
        if not isinstance(tape_barcode, str):
            raise TypeError(
                f"slk.{inspect.stack()[0][3]}: argument 'tape_barcode' needs to be 'str' but got "
                + f"'{type(tape_barcode).__name__}'."
            )
        slk_call.append("--tape-barcode")
        slk_call.append(tape_barcode)

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tape_library_raw(
    tape_id: Union[int, str, None] = None,
    tape_barcode: Union[str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Return the name of the library in which the tape is stored

    :param tape_id: id of a tape in the tape library
    :type tape_id: int or str or None
    :param tape_barcode: barcode of a tape in the tape library
    :type tape_barcode: str or None
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "tape_library"]

    if tape_id is not None:
        if not isinstance(tape_id, (int, str)):
            raise TypeError(
                f"slk.{inspect.stack()[0][3]}: argument 'tape_id' needs to be 'int' or 'str' but got "
                + f"'{type(tape_id).__name__}'."
            )
        slk_call.append(str(tape_id))
    if tape_barcode is not None:
        if not isinstance(tape_barcode, str):
            raise TypeError(
                f"slk.{inspect.stack()[0][3]}: argument 'tape_barcode' needs to be 'str' but got "
                + f"'{type(tape_barcode).__name__}'."
            )
        slk_call.append("--tape-barcode")
        slk_call.append(tape_barcode)

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def iscached_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    search_id: Union[str, int, None] = None,
    recursive: bool = False,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get info whether file is in HSM cache or not

    :param resource_path: resource (full path)
    :type resource_path: str or Path or None
    :param resource_id: a resource id
    :type resource_id: str or int or None
    :param search_id: id of a search
    :type search_id: int or str or None
    :param recursive: export metadata from all files in gns_path recursively
    :type recursive: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "iscached"]

    if resource_path is not None:
        if not isinstance(resource_path, (str, Path)):
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
                + f"{type(resource_id).__name__}"
            )
        slk_call.append(str(resource_path))
    if resource_id is not None:
        if not isinstance(resource_id, (str, int)):
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
                + f"{type(resource_id).__name__}"
            )
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))
    if search_id is not None:
        if not isinstance(search_id, (str, int)):
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
                + f"{type(resource_id).__name__}"
            )
        slk_call.append("--search-id")
        slk_call.append(str(search_id))
    if recursive:
        slk_call.append("-R")
    if verbose:
        slk_call.append("-v")
    if double_verbose:
        slk_call.append("-vv")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def is_on_tape_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    search_id: Union[str, int, None] = None,
    recursive: bool = False,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get info whether file is stored on a tape or not (independent on whether it is stored in the cache)

    :param resource_path: resource (full path)
    :type resource_path: str or Path or None
    :param resource_id: a resource id
    :type resource_id: str or int or None
    :param search_id: id of a search
    :type search_id: int or str or None
    :param recursive: export metadata from all files in resource_path recursively
    :type recursive: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_path is not None and not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
            + f"{type(resource_id).__name__}"
        )

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    if search_id is not None and not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    slk_call = [SLK_HELPERS, "is_on_tape"]

    if resource_path is not None:
        slk_call.append(str(resource_path))
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))
    if search_id is not None:
        slk_call.append("--search-id")
        slk_call.append(str(search_id))
    if recursive:
        slk_call.append("-R")
    if verbose:
        slk_call.append("-v")
    if double_verbose:
        slk_call.append("-vv")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def group_files_by_tape_raw(
    resource_paths: Union[str, list, Path, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[int, str, None] = None,
    search_query: Union[str, None] = None,
    destination: Union[Path, str, None] = None,
    recursive: bool = False,
    preserve_path: bool = True,
    print_tape_id: bool = False,
    print_tape_barcode: bool = False,
    print_tape_status: bool = False,
    print_resource_id: bool = False,
    count_files: bool = False,
    gen_search_query: bool = False,
    run_search_query: bool = False,
    set_max_tape_number_per_search: int = -1,
    json: bool = False,
    full: bool = False,
    details: bool = False,
    count_tapes: bool = False,
    evaluate_regex_in_input: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess, subprocess.Popen]:
    """groups a list of files based on which tape they are stored on

    Receives a list of files or a search id as input. Looks up which files are stored in the HSM-Cache and which are
    not stored in the HSM-Cache but only on tape. Files on tape are grouped by tape: each line of the output contains
    all files which are on one tape. The user can directly create a search query for retrieving all files from one
    tape (--gen-search-query) or directly run this search (--run-search-query). In the latter case, the search id is
    printed per tape. If the user wants to know the tape id and the tape status, she/he might use --print-tape-id and
    --print-tape-status, respectively.

    :param resource_paths: list of resource paths to be searched for
    :type resource_paths: str or list or Path-like or None
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: int or str or None
    :param search_query: a search query
    :type search_query: str
    :param destination: destination directory to check whether some of the needed files are already available and
        don't need to be retrieved
    :type destination: str or Path
    :param recursive: do recursive search in the namespaces
    :type recursive: bool
    :param preserve_path: preserve original file path, Default: True
    :type preserve_path: bool
    :param print_tape_id: print the tape id on the far left followed by a ':', Default: False
    :type print_tape_id: bool
    :param print_tape_barcode: print the tape barcode on the far left followed by a ':', Default: False
    :type print_tape_barcode: bool
    :param print_tape_status:  print the status ('avail' or 'blocked') of the tape of each file group; if print_tape_id
        is set, this is printed: 'TAPE_ID, TAPE_STATUS: FILES'; if this is not set, this is printed: TAPE_STATUS: FILES,
        Default: False
    :type print_tape_status: bool
    :param print_resource_id: print the resource id for each file instead of its path. Is ignored when
        '--gen-search-query', '--run-search-query', '--full' or '--count-files' are set. Default: False.
    :type print_resource_id: bool
    :param count_files: count files instead of printing file list
    :type count_files: bool
    :param gen_search_query: generate and print (a) search query strings instead of the lists of files per tape,
        Default: False
    :type gen_search_query: bool
    :param run_search_query: generate and run (a) search query strings instead of the lists of files per tape and print
        the search i, Default: False
    :type run_search_query: bool
    :param set_max_tape_number_per_search: number of tapes per search; if '-1' => the parameter is not set
    :type set_max_tape_number_per_search: int
    :param json: return the output as JSON
    :type json: bool
    :param full: print useful information and run search query per tape; implies print_tape_barcode, print_tape_status
        and run_search_query
    :type full: bool
    :param details: print useful information; implies print_tape_barcode and print_tape_status
    :type details: bool
    :param count_tapes: only count the number of involved tapes and print the number incl. some text
    :type count_tapes: bool
    :param evaluate_regex_in_input: expect that a regular expression in part of the input (file names) and evalute it
    :type evaluate_regex_in_input: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output), 3 (running subproc.)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess, subprocess.Popen]

    .. seealso::
        * :py:meth:`~pyslk.group_files_by_tape`

    """
    # TODO: make nice version
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = SLK_HELPERS + " group_files_by_tape"

    if resource_paths is not None:
        if isinstance(resource_paths, str) and resource_paths != "":
            slk_call = slk_call + " " + resource_paths
        elif isinstance(resource_paths, list) and len(resource_paths) > 0:
            if not all([isinstance(r, (str, Path)) for r in resource_paths]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: if argument 'resource_paths' is of type 'list' its items need to be"
                    + "of type 'str' or Path-like but got type(s): "
                    + f"'{', '.join([type(r).__name__ for r in resource_paths if not isinstance(r, (str, Path))])}'."
                )
            slk_call = slk_call + " " + " ".join([str(res) for res in resource_paths])
        elif isinstance(resource_paths, Path):
            slk_call = slk_call + " " + str(resource_paths)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_paths" has to be of type "str", path-like or "list" but is '
                + f"{type(resource_paths).__name__}"
            )
    if resource_ids is not None:
        if isinstance(resource_ids, str) and resource_ids != "":
            slk_call = slk_call + " --resource-ids " + resource_ids
        elif isinstance(resource_ids, int):
            slk_call = slk_call + " " + str(resource_ids)
        elif isinstance(resource_ids, list) and len(resource_ids) > 0:
            if not all([isinstance(r, (str, int)) for r in resource_ids]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: if argument 'resource_ids' is of type 'list' its items need to be"
                    + "of type 'str' or 'int' but got type(s): "
                    + f"'{', '.join([type(r).__name__ for r in resource_ids if not isinstance(r, (str, Path))])}'."
                )
            slk_call = slk_call + " " + " ".join([str(res) for res in resource_ids])
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_paths" has to be of type "str", path-like or "list" but is '
                + f"{type(resource_paths).__name__}"
            )
    if search_id is not None and not (isinstance(search_id, int) and search_id == -1):
        slk_call = slk_call + " --search-id " + str(search_id)
    if search_query is not None:
        if isinstance(search_query, str):
            slk_call = slk_call + " --search-query " + f"'{search_query}'"
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "search_query" has to be of type "str" but is '
                + f"{type(search_query).__name__}"
            )

    if destination is not None:
        if isinstance(destination, (str, Path)):
            slk_call = slk_call + " -dst " + str(destination)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'destination' needs to be 'str' or path-like but got "
                + f"'{type(destination).__name__}'"
            )
    if recursive:
        slk_call = slk_call + " --recursive"
    if preserve_path:
        slk_call = slk_call + " -ns"
    if print_tape_id:
        slk_call = slk_call + " --print-tape-id"
    if print_tape_barcode:
        slk_call = slk_call + " --print-tape-barcode"
    if print_tape_status:
        slk_call = slk_call + " --print-tape-status"
    if print_resource_id:
        slk_call = slk_call + " --print-resource-id"
    if count_files:
        slk_call = slk_call + " --count-files"
    if gen_search_query:
        slk_call = slk_call + " --gen-search-query"
    if run_search_query:
        slk_call = slk_call + " --run-search-query"
    if set_max_tape_number_per_search != -1:
        slk_call = (
            slk_call
            + " --set-max-tape-number-per-search "
            + str(set_max_tape_number_per_search)
        )
    if json:
        slk_call = slk_call + " --json"
    if full:
        slk_call = slk_call + " --full"
    if details:
        slk_call = slk_call + " --details"
    if count_tapes:
        slk_call = slk_call + " --count-tapes"
    if evaluate_regex_in_input:
        slk_call = slk_call + " --evaluate-regex-in-input"

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def print_rcrs_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    json: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """prints resource content record (rcr) information

    :param resource_path: resource (full path)
    :type resource_path: str or Path or None
    :param resource_id: a resource id
    :type resource_id: str or int or None
    :param json: print json
    :type json: ``bool``
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "print_rcrs"]

    if resource_path is not None:
        if not isinstance(resource_path, (str, Path)):
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
                + f"{type(resource_id).__name__}"
            )
        slk_call.append(str(resource_path))
    if resource_id is not None:
        if not isinstance(resource_id, (str, int)):
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
                + f"{type(resource_id).__name__}"
            )
        if isinstance(resource_id, str):
            try:
                int(resource_id)
            except ValueError:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: 'resource_id' cannot be processed; need 'str', which holds an "
                    + f"integer number, or 'int' but got '{type(resource_id).__name__}' with value '{resource_id}'"
                )
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))

    # --json
    if not isinstance(json, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json' has wrong type; need 'bool' but got "
            + f"'{type(json).__name__}'"
        )
    if json:
        slk_call.append("--json")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import datetime
import re
from pathlib import Path

import pytest

from pyslk import (
    StatusJob,
    is_job_finished,
    is_job_successful,
)
from pyslk import list as ls
from pyslk.base.login import (
    expiration_date,
    hostname,
    session,
    valid_session,
    valid_token,
)
from pyslk.base.resource_path import get_resource_path
from pyslk.base.searching import search, total_number_search_results
from pyslk.base.versioning import cli_versions, version_slk, version_slk_helpers
from pyslk.core.gfbt import count_tapes, group_files_by_tape
from pyslk.core.resource import (
    get_resource_id,
    get_resource_permissions,
    get_resource_size,
    has_no_flag_partial,
    has_no_flag_partial_details,
)
from pyslk.core.resource_extras import (
    get_resource_type,
    is_file,
    is_namespace,
    resource_exists,
)
from pyslk.core.storage import (
    get_tape_barcode,
    get_tape_id,
    is_cached,
    is_cached_details,
    tape_exists,
    tape_status,
)
from pyslk.jobs import (
    get_job_status,
    is_job_processing,
    is_job_queued,
    job_exists,
    job_queue,
)
from pyslk.pyslk_exceptions import PySlkException


class TestBasics:
    def test_valid_session(self):
        assert valid_session()

    def test_valid_token(self):
        assert valid_token()

    def test_hostname(self):
        output = hostname()
        assert isinstance(output, str)
        assert len(output) > 0

    def test_expiration_date(self):
        output = expiration_date()
        now = datetime.datetime.now(tz=output.tzinfo)
        in_one_year = datetime.datetime.now(tz=output.tzinfo)
        in_one_year = in_one_year.replace(year=in_one_year.year + 1)
        assert isinstance(output, datetime.datetime)
        assert output > now
        assert output < in_one_year

    def test_session(self):
        output = session()
        now = datetime.datetime.now(tz=output.tzinfo)
        in_one_year = datetime.datetime.now(tz=output.tzinfo)
        in_one_year = in_one_year.replace(year=in_one_year.year + 1)
        assert isinstance(output, datetime.datetime)
        assert output > now
        assert output < in_one_year


class TestBasicOperations:
    def test_mkdir_rename_move_delete(self):
        pass

    def test_mkdir_archive_rename_move_detele(self):
        pass


class TestVersionFunctions:
    def test_version_slk_helpers(self):
        slk_helpers_version = re.findall(
            "[0-9]+[.][0-9]+[.][0-9]+", version_slk_helpers()
        )
        assert len(slk_helpers_version) == 1

    def test_version_slk(self):
        slk_version = re.findall("[0-9]+[.][0-9]+[.][0-9]+", version_slk())
        assert len(slk_version) == 1

    @pytest.mark.parametrize(
        "test_input,in_output,not_output",
        [
            (None, ["slk", "slk_helpers"], list()),
            ("slk", ["slk"], ["slk_helpers"]),
            ("slk_helpers", ["slk_helpers"], ["slk"]),
            ("slk", ["slk"], ["slk_helpers"]),
            (["slk"], ["slk"], ["slk_helpers"]),
            (["slk", "slk_helpers"], ["slk", "slk_helpers"], list()),
            (("slk", "slk_helpers"), ["slk", "slk_helpers"], list()),
        ],
    )
    def test_cli_versions(self, test_input, in_output, not_output):
        slk_versions = cli_versions(test_input)
        for i in in_output:
            assert i in slk_versions
            slk_version = re.findall("[0-9]+[.][0-9]+[.][0-9]+", version_slk())
            assert len(slk_version) == 1
        for i in not_output:
            assert i not in slk_versions


class TestListFunctions:
    def test_ls_basic(self):
        # path for test
        resource_path = "/arch/bm0146/k204221/iow"
        # things to compare
        columns = [
            "permissions",
            "owner",
            "group",
            "filesize",
            "date",
            "time",
            "filename",
        ]
        n_rows = 23
        files = [
            "/arch/bm0146/k204221/iow/INDEX.txt",
            "/arch/bm0146/k204221/iow/iow_data2_001.tar",
            "/arch/bm0146/k204221/iow/iow_data2_002.tar",
            "/arch/bm0146/k204221/iow/iow_data2_003.tar",
            "/arch/bm0146/k204221/iow/iow_data2_004.tar",
            "/arch/bm0146/k204221/iow/iow_data2_005.tar",
            "/arch/bm0146/k204221/iow/iow_data2_006.tar",
            "/arch/bm0146/k204221/iow/iow_data3_001.tar",
            "/arch/bm0146/k204221/iow/iow_data3_002.tar",
            "/arch/bm0146/k204221/iow/iow_data4_001.tar",
            "/arch/bm0146/k204221/iow/iow_data4_002.tar",
            "/arch/bm0146/k204221/iow/iow_data5_001.tar",
            "/arch/bm0146/k204221/iow/iow_data5_002.tar",
            "/arch/bm0146/k204221/iow/iow_data5_003.tar",
            "/arch/bm0146/k204221/iow/iow_data5_004.tar",
            "/arch/bm0146/k204221/iow/iow_data5_005.tar",
            "/arch/bm0146/k204221/iow/iow_data5_006.tar",
            "/arch/bm0146/k204221/iow/iow_data_001.tar",
            "/arch/bm0146/k204221/iow/iow_data_002.tar",
            "/arch/bm0146/k204221/iow/iow_data_003.tar",
            "/arch/bm0146/k204221/iow/iow_data_004.tar",
            "/arch/bm0146/k204221/iow/iow_data_005.tar",
            "/arch/bm0146/k204221/iow/iow_data_006.tar",
        ]

        # call function
        output = ls(resource_path)

        # compare output
        assert list(output.columns) == columns
        assert len(output) == n_rows
        for f in files:
            assert f in list(output["filename"])
        for f in list(output["filename"]):
            assert f in files


class TestResourceFunctions:
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", True, None),
            (Path("/arch/bm0146/k204221/iow"), True, None),
            ("/arch/bm0146/k204221/not_exists", False, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", False, None),
            (49058658013, True, None),
            ("49058658013", True, None),
            (49058705497, True, None),
            (999999999999999, False, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_resource_exists(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert resource_exists(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                resource_exists(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", None, None),
            (Path("/arch/bm0146/k204221/iow"), None, TypeError),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", None, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, Path("/arch/bm0146/k204221/iow"), None),
            ("49058658013", Path("/arch/bm0146/k204221/iow"), None),
            (49058705497, Path("/arch/bm0146/k204221/iow/INDEX.txt"), None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_path(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_resource_path(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_path(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", 49058658013, None),
            (Path("/arch/bm0146/k204221/iow"), 49058658013, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", 49058705497, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, None, TypeError),
            ("49058658013", None, None),
            (49058705497, None, TypeError),
            (999999999999999, None, TypeError),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_id(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_resource_id(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_id(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", "NAMESPACE", None),
            (Path("/arch/bm0146/k204221/iow"), "NAMESPACE", None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", "FILE", None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, "NAMESPACE", None),
            ("49058658013", "NAMESPACE", None),
            (49058705497, "FILE", None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_type(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_resource_type(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_type(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", 0, None),
            (Path("/arch/bm0146/k204221/iow"), 0, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", 1268945, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, 0, None),
            ("49058658013", 0, None),
            (49058705497, 1268945, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_size(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_resource_size(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_size(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", False, None),
            (Path("/arch/bm0146/k204221/iow"), False, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, False, None),
            ("49058658013", False, None),
            (49058705497, True, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_is_file(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_file(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_file(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", True, None),
            (Path("/arch/bm0146/k204221/iow"), True, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", False, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, True, None),
            ("49058658013", True, None),
            (49058705497, False, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_is_namespace(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_namespace(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_namespace(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (
                "/arch/bm0146/k204221/iow",
                "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$",
                None,
            ),
            (
                Path("/arch/bm0146/k204221/iow"),
                "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$",
                None,
            ),
            ("/arch/bm0146/k204221/not_exists", None, None),
            (
                "/arch/bm0146/k204221/iow/INDEX.txt",
                "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$",
                None,
            ),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$", None),
            ("49058658013", "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$", None),
            (49058705497, "^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$", None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_permissions_text(
        self, test_input, expected_output, expected_error
    ):
        if expected_error is None:
            # expect no error
            output = get_resource_permissions(test_input)
            if expected_output is None:
                assert output == expected_output
            else:
                assert len(re.findall(expected_output, output)) == 1
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_permissions(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", "^[01234567]{3}$", None),
            (Path("/arch/bm0146/k204221/iow"), "^[01234567]{3}$", None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", "^[01234567]{3}$", None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, "^[01234567]{3}$", None),
            ("49058658013", "^[01234567]{3}$", None),
            (49058705497, "^[01234567]{3}$", None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], None, TypeError),
        ],
    )
    def test_get_resource_permissions_octal(
        self, test_input, expected_output, expected_error
    ):
        if expected_error is None:
            # expect no error
            output = get_resource_permissions(test_input, as_octal_number=True)
            if expected_output is None:
                assert output == expected_output
            else:
                assert isinstance(output, str)
                assert len(re.findall(expected_output, output)) == 1
        else:
            # expect error
            with pytest.raises(expected_error):
                get_resource_permissions(test_input)


class TestTapeFunctions:
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("M24361M8", 136343, None),
            ("M24361M1", None, None),
            (12345, None, TypeError),
        ],
    )
    def test_get_tape_id(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_tape_id(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_tape_id(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (136343, "M24361M8", None),
            (99999999, None, None),
            ("M24361M8", None, TypeError),
        ],
    )
    def test_get_tape_barcode(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_tape_barcode(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_tape_barcode(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("M24361M8", True, None),
            ("M24361M1", False, None),
            (136343, True, None),
            (99999999, False, None),
            (list("M24361M8"), None, TypeError),
        ],
    )
    def test_tape_exists(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert tape_exists(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                tape_exists(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("M24361M8", "AVAILABLE", None),
            ("M24361M1", None, None),
            (136343, "AVAILABLE", None),
            (99999999, None, None),
            (list("M24361M8"), None, TypeError),
        ],
    )
    def test_tape_status(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert tape_status(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                tape_status(test_input)


class TestJobFunctions:
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, True, None),
            (133873, True, None),
            (99999999, False, None),
            ("abc", None, TypeError),
        ],
    )
    def test_job_exists(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert job_exists(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                job_exists(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, StatusJob("FAILED"), None),
            (133873, StatusJob("SUCCESSFUL"), None),
            (99999999, None, None),
            ("abc", None, TypeError),
        ],
    )
    def test_job_status(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert get_job_status(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                get_job_status(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, False, None),
            (133873, False, None),
            (99999999, None, None),
            ("abc", None, TypeError),
        ],
    )
    def test_is_job_queued(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_job_queued(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_job_queued(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, False, None),
            (133873, False, None),
            (99999999, None, None),
            ("abc", None, TypeError),
        ],
    )
    def test_is_job_processing(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_job_processing(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_job_processing(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, True, None),
            (133873, True, None),
            (99999999, None, None),
            ("abc", None, TypeError),
        ],
    )
    def test_is_job_finished(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_job_finished(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_job_finished(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (133874, False, None),
            (133873, True, None),
            (99999999, None, None),
            ("abc", None, TypeError),
        ],
    )
    def test_is_job_successful(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_job_successful(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_job_successful(test_input)

    def test_job_queue(self):
        assert job_queue() is not None


class TestGFBT:
    # @pytest.mark.new
    @pytest.mark.parametrize(
        "test_input,expected_error",
        [
            (("/arch/bm0146/k204221/iow", None, None), None),
            ((None, None, None), ValueError),
            (("/arch/bm0146/k204221/iow", 3, None), ValueError),
            (
                (
                    "/arch/bm0146/k204221/iow",
                    None,
                    '{"path":{"$gte":"/arch/bm0146/k204221/iow"}}',
                ),
                ValueError,
            ),
            ((None, 3, '{"path":{"$gte":"/arch/bm0146/k204221/iow"}}'), ValueError),
            ((1, None, None), TypeError),
            ((None, "abc", None), ValueError),
            ((None, None, 3), TypeError),
            (("abc", None, None), FileNotFoundError),
        ],
    )
    # comment on test: ((None, "abc", None), ValueError),
    #   We throw ValueError and no TypeError because we accept strings for this
    #   argument and try to convert them to int. If the argument was "4", the
    #   would be not error.
    def test_count_tapes_0(self, test_input, expected_error):
        if expected_error is None:
            # expect no error; test whether it works at all
            count_tapes(
                resource_path=test_input[0],
                search_id=test_input[1],
                search_query=test_input[2],
                recursive=True,
            )
            assert True
        else:
            # expect error
            with pytest.raises(expected_error):
                count_tapes(
                    resource_path=test_input[0],
                    search_id=test_input[1],
                    search_query=test_input[2],
                    recursive=True,
                )

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", (12, 0), None),
            ("/arch/bm0146/k204221/iow/iow_data2_001.tar", (1, 0), None),
            (Path("/arch/bm0146/k204221/iow"), (12, 0), None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", (0, 0), None),
            (
                [
                    "/arch/bm0146/k204221/iow/iow_data_001.tar",
                    "/arch/bm0146/k204221/iow/iow_data2_002.tar",
                    "/arch/bm0146/k204221/iow/iow_data5_002.tar",
                ],
                (2, 0),
                None,
            ),
            (
                [
                    "/arch/bm0146/k204221/iow/iow_data_006.tar",
                    "/arch/bm0146/k204221/iow/iow_data5_006.tar",
                    "/arch/bm0146/k204221/iow/iow_data5_002.tar",
                ],
                (2, 0),
                None,
            ),
            ({"abc": 4}, None, TypeError),
        ],
    )
    def test_count_tapes_1(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            output_dict = count_tapes(resource_path=test_input, recursive=True)
            if expected_output[0] == 12:
                # in case that some files were recalled for other tests
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    <= expected_output[0]
                ) and (
                    output_dict.get("n_tapes__single_tape_files", None)
                    >= (expected_output[0] - 2)
                )
            else:
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    == expected_output[0]
                )
            assert (
                output_dict.get("n_tapes__multi_tape_files", None) == expected_output[1]
            )
        else:
            # expect error
            with pytest.raises(expected_error):
                count_tapes(test_input, recursive=True)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ('{"path":{"$gte":"/arch/bm0146/k204221/iow"}}', (12, 0), None),
            (
                '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},'
                + '{"resources.name":{"$regex":"iow_data2_001.tar"}}]}',
                (1, 0),
                None,
            ),
            (
                '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},'
                + '{"resources.name":{"$regex":"INDEX.txt"}}]}',
                (0, 0),
                None,
            ),
            (
                '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},'
                + '{"resources.name":'
                + '{"$regex":"iow_data2_001.tar|iow_data2_002.tar|iow_data5_002.tar"}}]}',
                (3, 0),
                None,
            ),
            (
                '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},'
                + '{"resources.name":'
                + '{"$regex":"iow_data_006.tar|iow_data5_006.tar|iow_data5_002.tar"}}]}',
                (2, 0),
                None,
            ),
            ((1), None, TypeError),
        ],
    )
    def test_count_tapes_2(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            # run with search query
            output_dict = count_tapes(search_query=test_input, recursive=True)
            if expected_output[0] == 12:
                # in case that some files were recalled for other tests
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    <= expected_output[0]
                ) and (
                    output_dict.get("n_tapes__single_tape_files", None)
                    >= (expected_output[0] - 2)
                )
            else:
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    == expected_output[0]
                )
            assert (
                output_dict.get("n_tapes__multi_tape_files", None) == expected_output[1]
            )
            # perform search and run with search id
            search_id = search(test_input)
            output_dict = count_tapes(search_id=search_id, recursive=True)
            if expected_output[0] == 12:
                # in case that some files were recalled for other tests
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    <= expected_output[0]
                ) and (
                    output_dict.get("n_tapes__single_tape_files", None)
                    >= (expected_output[0] - 2)
                )
            else:
                assert (
                    output_dict.get("n_tapes__single_tape_files", None)
                    == expected_output[0]
                )
            assert (
                output_dict.get("n_tapes__multi_tape_files", None) == expected_output[1]
            )
        else:
            # expect error
            with pytest.raises(expected_error):
                count_tapes(resource_path=test_input, recursive=True)

    # @pytest.mark.new
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (("/arch/bm0146/k204221/iow3", None, None), (3, 0), None),
            ((Path("/arch/bm0146/k204221/iow3"), None, None), (3, 0), None),
            (("/arch/bm0146/k204221/not_exists", None, None), None, FileNotFoundError),
            (
                (
                    "/arch/bm0146/k204221/iow3/ocean_day3d_phoswam_v04_ist_2006.nc",
                    None,
                    None,
                ),
                (1, 0),
                None,
            ),
            (("/arch/bm0146/k204221/iow/INDEX.txt", None, None), (0, 0), None),
            (
                ("/arch/bm0146/k204221/iow3/not_exists.txt", None, None),
                None,
                FileNotFoundError,
            ),
            (
                (None, None, '{"path": {"$gte": "/arch/bm0146/k204221/iow3"}}'),
                (3, 0),
                None,
            ),
            ((None, 923792, None), (3, 0), None),
            ((None, "923792", None), (3, 0), None),
            (
                (
                    None,
                    None,
                    '{"$and": [{"path": {"$gte": "/arch/bd1022/b309073/Res"}}, '
                    + '{"resources.name": {"$regex": "[12][09][09]._(tracer_gp|vaxtra)"}}]}',
                ),
                (82, 13),
                None,
            ),
            ((None, 1.1, None), None, TypeError),
            ((None, None, 123456), None, TypeError),
            ((None, 451086, None), None, PySlkException),
            ((["/arch/bm0146/k204221/iow3"], None, None), (3, 0), None),
            ((12345, None, None), None, TypeError),
        ],
    )
    def test_group_files_by_tape_count(
        self, test_input, expected_output, expected_error
    ):
        if expected_error is None:
            # expect no error
            output = count_tapes(
                resource_path=test_input[0],
                search_id=test_input[1],
                search_query=test_input[2],
                recursive=True,
            )
            assert "n_tapes__single_tape_files" in output
            assert "n_tapes__multi_tape_files" in output
            assert output["n_tapes__single_tape_files"] == expected_output[0]
            assert output["n_tapes__multi_tape_files"] == expected_output[1]
        else:
            # expect error
            with pytest.raises(expected_error):
                count_tapes(
                    resource_path=test_input[0],
                    search_id=test_input[1],
                    search_query=test_input[2],
                    recursive=True,
                )

    # @pytest.mark.new
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            (("/arch/bm0146/k204221/iow3", None, None), (3, "B09384L5"), None),
            ((Path("/arch/bm0146/k204221/iow3"), None, None), (3, "B09384L5"), None),
            (("/arch/bm0146/k204221/not_exists", None, None), None, FileNotFoundError),
            (
                (
                    "/arch/bm0146/k204221/iow3/ocean_day3d_phoswam_v04_ist_2006.nc",
                    None,
                    None,
                ),
                (1, "B09398L5"),
                None,
            ),
            (
                ("/arch/bm0146/k204221/iow3/not_exists.txt", None, None),
                None,
                FileNotFoundError,
            ),
            (
                (None, None, '{"path": {"$gte": "/arch/bm0146/k204221/iow3"}}'),
                (3, "B09384L5"),
                None,
            ),
            ((None, 923792, None), (3, "B09384L5"), None),
            ((None, "923792", None), (3, "B09384L5"), None),
            (
                (
                    None,
                    None,
                    '{"$and": [{"path": {"$gte": "/arch/bd1022/b309073/Res"}}, '
                    + '{"resources.name": {"$regex": "[12][09][09]._(tracer_gp|vaxtra)"}}]}',
                ),
                (83, "M10397M8"),
                None,
            ),
            ((None, 1.1, None), None, TypeError),
            ((None, None, 123456), None, TypeError),
            ((None, 451086, None), None, PySlkException),
            ((["/arch/bm0146/k204221/iow3"], None, None), (3, "B09384L5"), None),
            ((12345, None, None), None, TypeError),
        ],
    )
    def test_group_files_by_tape_details(
        self, test_input, expected_output, expected_error
    ):
        expected_keys = [
            "barcode",
            "description",
            "file_count",
            "files",
            "id",
            "location",
            "search_query",
            "status",
        ]
        if expected_error is None:
            # expect no error
            output = group_files_by_tape(
                resource_path=test_input[0],
                search_id=test_input[1],
                search_query=test_input[2],
                recursive=True,
            )
            assert len(output) == expected_output[0]
            assert all([key in tape for key in expected_keys for tape in output])
            assert any(
                [tape.get("barcode", None) == expected_output[1] for tape in output]
            )
        else:
            # expect error
            with pytest.raises(expected_error):
                group_files_by_tape(
                    resource_path=test_input[0],
                    search_id=test_input[1],
                    search_query=test_input[2],
                    recursive=True,
                )


# {"path":{"$gte":"/arch/bm0146/k204221/iow"}}


class TestStorageInfoFunctions:
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", True, None),
            (Path("/arch/bm0146/k204221/iow"), True, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, True, None),
            ("49058658013", True, None),
            (49058705497, True, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], True, None),
        ],
    )
    def test_has_no_flag_partial(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert has_no_flag_partial(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                has_no_flag_partial(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", True, None),
            (Path("/arch/bm0146/k204221/iow"), True, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, True, None),
            ("49058658013", True, None),
            (49058705497, True, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], True, None),
        ],
    )
    def test_has_no_flag_partial_details(
        self, test_input, expected_output, expected_error
    ):
        if expected_error is None:
            # expect no error
            output = has_no_flag_partial_details(test_input)
            if expected_output is None:
                assert output == expected_output
            else:
                assert isinstance(output, dict)
                assert "no_flag_partial" in output
                assert "flag_partial" in output
                assert (
                    Path("/arch/bm0146/k204221/iow/INDEX.txt")
                    in output["no_flag_partial"]
                )
        else:
            # expect error
            with pytest.raises(expected_error):
                has_no_flag_partial_details(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("/arch/bm0146/k204221/iow", False, None),
            (Path("/arch/bm0146/k204221/iow"), False, None),
            ("/arch/bm0146/k204221/not_exists", None, None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, None),
            ("/arch/bm0146/k204221/iow/iow_data_006.tar", False, None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, None),
            (49058658013, False, None),
            ("49058658013", False, None),
            (49058705497, True, None),
            (999999999999999, None, None),
            (3.2, None, TypeError),
            (["/arch/bm0146/k204221/iow"], False, None),
        ],
    )
    def test_is_cached(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_cached(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_cached(test_input)

    # will fail due to issue #40 in the slk_helpers
    @pytest.mark.xfail
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [(49058658013, False, None), ("49058658013", False, None)],
    )
    def test_is_cached_fail(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            assert is_cached(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                is_cached(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,check_files,expected_error",
        [
            ("/arch/bm0146/k204221/iow", True, (True, True), None),
            (Path("/arch/bm0146/k204221/iow"), True, (True, True), None),
            ("/arch/bm0146/k204221/not_exists", None, (False, False), None),
            ("/arch/bm0146/k204221/iow/INDEX.txt", True, (True, False), None),
            ("/arch/bm0146/k204221/iow/iow_data_006.tar", False, (False, True), None),
            ("/arch/bm0146/k204221/iow/not_exists.txt", None, (False, False), None),
            (49058658013, True, (True, True), None),
            ("49058658013", True, (True, True), None),
            (49058705497, True, (True, False), None),
            (999999999999999, None, (False, False), None),
            (3.2, None, (False, False), TypeError),
            (["/arch/bm0146/k204221/iow"], True, (True, True), None),
        ],
    )
    def test_is_cached_details(
        self, test_input, expected_output, check_files, expected_error
    ):
        if expected_error is None:
            # expect no error
            output = is_cached_details(test_input)
            if expected_output is None:
                assert output == expected_output
            else:
                print(output)
                assert isinstance(output, dict)
                assert "cached" in output
                assert "not_cached" in output
                if check_files[0]:
                    assert (
                        Path("/arch/bm0146/k204221/iow/INDEX.txt") in output["cached"]
                    )
                if check_files[1]:
                    assert (
                        Path("/arch/bm0146/k204221/iow/iow_data_006.tar")
                        in output["not_cached"]
                    )
                # assert "/arch/bm0146/k204221/iow/INDEX.txt" in [str(out) for out in output["cached"]]
                # assert "/arch/bm0146/k204221/iow/iow_data_006.tar" in [str(out) for out in output["not_cached"]]
        else:
            # expect error
            with pytest.raises(expected_error):
                is_cached_details(test_input)


class TestSearch:
    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ('{"path": {"$gte": "/arch/bm0146/k204221"}}}', True, None),
            (123456, False, TypeError),
            ('{"$gte": "/arch/bm0146/k204221/cera_test"}}', False, PySlkException),
            (
                '{"path": {"$gte": "/arch/bm0146/k204221/abcdefgh"}}}',
                False,
                PySlkException,
            ),
        ],
    )
    def test_search_basic(self, test_input, expected_output, expected_error):
        if expected_error is None:
            # expect no error
            search_id_matches = re.findall("^[0-9]{1,7}$", str(search(test_input)))
            assert len(search_id_matches) == 1
        else:
            # expect error
            with pytest.raises(expected_error):
                search(test_input)

    @pytest.mark.parametrize(
        "test_input,expected_output,expected_error",
        [
            ("451901", 3, None),
            (451901, 3, None),
            (451900, 4, None),
            ("abc", None, ValueError),
            (1.2, None, TypeError),
            (99999999, None, None),
        ],
    )
    def test_total_number_search_results(
        self, test_input, expected_output, expected_error
    ):
        if expected_error is None:
            # expect no error
            assert total_number_search_results(test_input) == expected_output
        else:
            # expect error
            with pytest.raises(expected_error):
                total_number_search_results(test_input)


class TestChModGGrpOwn:
    def test_chgrp(self):
        # import pyslk
        # pyslk.chgrp("/arch/bm0146/k204221/iow", recursive=True, group="ka1209")
        # pyslk.chgrp("/arch/bm0146/k204221/iow", recursive=True, group="bm0146")
        # pyslk.chgrp("/arch/bm0146/k204221/iow", recursive=True, group="ka1209")
        # pyslk.chgrp("/arch/bm0146/k204221/iow", recursive=True, group=1076)
        pass

    def test_chown(self):
        # import pyslk
        # pyslk.chown("/arch/bm0146/k204221/iow", recursive=True, owner="k202147")
        # pyslk.chown("/arch/bm0146/k204221/iow", recursive=True, owner="k204221")
        # pyslk.chown("/arch/bm0146/k204221/iow", recursive=True, owner=200533)
        # pyslk.chown("/arch/bm0146/k204221/iow", recursive=True, owner=25301)
        pass



import os

import pytest

from pyslk.utils import (
    _parse_dates,
    _parse_list_to_rows,
    _parse_sizes,
    _rows_to_dataframe,
)

test_dir = os.path.dirname((os.path.abspath(__file__)))


@pytest.fixture
def test_output():
    filename = os.path.join(test_dir, "test_dir_output.txt")
    with open(filename) as f:
        test_str = f.read()
    return test_str


@pytest.fixture
def test_output_recursive():
    filename = os.path.join(test_dir, "test_recursive_output.txt")
    with open(filename) as f:
        test_str = f.read()
    return test_str


@pytest.fixture
def test_output_wildcard():
    filename = os.path.join(test_dir, "wildcard_output.txt")
    with open(filename) as f:
        test_str = f.read()
    return test_str


def test_wildcard_output(test_output_wildcard):
    rows = _parse_list_to_rows(
        test_output_wildcard.splitlines(),
        "/arch/ch0636/remo-runs/065002/1960/e065002t*.tar",
    )
    print(rows)
    df = _rows_to_dataframe(rows)
    files = list(df.filename)
    expect = [
        "/arch/ch0636/remo-runs/065002/1960/e065002t196001.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196002.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196003.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196004.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196005.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196006.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196007.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196008.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196009.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196010.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196011.tar",
        "/arch/ch0636/remo-runs/065002/1960/e065002t196012.tar",
    ]
    assert files == expect

    # test single
    raw = "-rw-r--r--- stronglink  0               1.0G   21 Jan 2022 18:57 ingest_01_1\nFiles: 1\n\x1b[?25h"
    result = _parse_list_to_rows(raw.splitlines(), "/test/test3/ingest_01_1")
    expect = [
        [
            "-rw-r--r---",
            "stronglink",
            "0",
            "1.0G",
            "21",
            "Jan",
            "2022",
            "18:57",
            "/test/test3/ingest_01_1",
        ]
    ]
    assert result == expect


def test_parse_to_dataframe(test_output):
    rows = _parse_list_to_rows(test_output.split("\n"))
    df = _rows_to_dataframe(rows)
    assert df.loc[273].filename == "uu0808"


def test_parse_to_dataframe_recursive(test_output_recursive):
    rows = _parse_list_to_rows(test_output_recursive.split("\n"))
    df = _rows_to_dataframe(rows)
    df = _parse_sizes(df)
    df = _parse_dates(df)
    assert df.filesize.sum() == 92119300000000.0



import inspect
import subprocess
from pathlib import Path
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK_HELPERS
from pyslk.utils import run_slk, which

__all__ = [
    "result_verify_job_raw",
    "submit_verify_job_raw",
]


def submit_verify_job_raw(
    resources: Union[str, Path, list[str], list[Path], None] = None,
    resource_ids: Union[int, str, list[int], list[str], None] = None,
    search_id: Union[int, str, None] = None,
    search_query: Union[str, None] = None,
    json_input_file: Union[str, Path, None] = None,
    recursive: bool = False,
    resume_on_page: int = 1,
    results_per_page: int = -1,
    end_on_page: int = -1,
    save_mode: bool = False,
    verbose: bool = False,
    json: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess, subprocess.Popen]:
    """
    submit a verify job

    :param resources: provide a list of resources over which the verify job should be run
    :type resources: `str` or `Path` or `list[str]` or `list[Path]` or `None`
    :param resource_ids: provide a list of ids of resources over which the verify job should be run
    :type resource_ids: `str` or `int` or `list[str]` or `list[int]` or `None`
    :param search_id: provide a search id pointing to files over which the verify job should be run
    :type search_id: `str` or `int` or `None`
    :param search_query: provide a search query to select files over which the verify job should be run
    :type search_query: `str` or `None`
    :param json_input_file: read a JSON file which contains payload of the verify job (admin users only)
    :type json_input_file: `str` or `Path`  or `None`
    :param recursive: iterate namespaces recursively
    :type recursive: `bool`
    :param resume_on_page: resume the command and start with search result page `resume_on_page`; internally, 1000
            search results are on one 'page' and fetched by one request; you do not necessarily have read permissions
            for all of these files
    :type resume_on_page: `int`
    :param results_per_page: number of search results requested per page
    :type results_per_page: `int`
    :param end_on_page: end with search result page `end_on_page`; internally, 1000 search results
            are on one 'page' and fetched by one request; you do not necessarily have read permissions for all of
            these files
    :type end_on_page: `int`
    :param save_mode: save mode suggested to be used in times of many timeouts; please do not regularly use this
            parameter; start one verify job per page of search results instead of one verify job for 50 pages of
            search results
    :type save_mode: `bool`
    :param verbose: verbose output
    :type verbose: `bool`
    :param json: print output as JSON
    :type json: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :return: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = SLK_HELPERS + " submit_verify_job"

    if resources is not None:
        if isinstance(resources, str) and resources != "":
            slk_call = slk_call + " " + resources
        elif isinstance(resources, list) and len(resources) > 0:
            if not all([isinstance(r, (str, Path)) for r in resources]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: if argument 'resources' is of type 'list' its items need to be"
                    + "of type 'str' or Path-like but got type(s): "
                    + f"'{', '.join([type(r).__name__ for r in resources if not isinstance(r, (str, Path))])}'."
                )
            slk_call = slk_call + " " + " ".join([str(res) for res in resources])
        elif isinstance(resources, Path):
            slk_call = slk_call + " " + str(resources)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resources" has to be of type "str", path-like or "list" but is '
                + f"{type(resources).__name__}"
            )
    if resource_ids is not None:
        if isinstance(resource_ids, str) and resource_ids != "":
            slk_call = slk_call + " --resource-ids " + resource_ids
        elif isinstance(resource_ids, int):
            slk_call = slk_call + " --resource-ids " + str(resource_ids)
        elif isinstance(resource_ids, list) and len(resources) > 0:
            if not all([isinstance(r, (str, int)) for r in resources]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: if argument 'resource_ids' is of type 'list' its items need to be"
                    + "of type 'str' or 'int' but got type(s): "
                    + f"'{', '.join([type(r).__name__ for r in resources if not isinstance(r, (str, int))])}'."
                )
            slk_call = (
                slk_call
                + " --resource-ids "
                + " ".join([str(res) for res in resources])
            )
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resources" has to be of type "str", path-like or "list" but is '
                + f"{type(resources).__name__}"
            )

    # --search-id
    if search_id is not None:
        if isinstance(search_id, int):
            if search_id > 0:
                # correct search id provided
                slk_call = slk_call + " --search-id " + str(search_id)
            elif search_id < -2:
                # wrong search id provided
                raise ValueError(
                    f"pyslk.{inspect.stack()[0][3]}: argument 'search_id' needs to be > -2 but is {search_id} ('> 0': "
                    + "considered as normal search id; '0' and '-1': considered as None)."
                )
            # case search_id == 0 or -1: assume that it means the same as None
        elif isinstance(search_id, str):
            slk_call = slk_call + " --search-id " + search_id
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_id' has to be of type 'str' or 'int' but is "
                + f"{type(search_id).__name__}"
            )

    # --search-query
    if search_query is not None:
        if isinstance(search_query, str):
            slk_call = slk_call + " --search-query " + f"'{search_query}'"
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_query' has wrong type; need 'str' but got "
                + f"'{type(search_query).__name__}'"
            )

    # --infile
    if json_input_file is not None:
        if isinstance(json_input_file, (str, Path)):
            slk_call = slk_call + " --infile " + str(json_input_file)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'json_input_file' has wrong type; need 'str' or 'Path' but got "
                + f"'{type(json_input_file).__name__}'"
            )

    # --resume-on-page
    if not isinstance(resume_on_page, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resume_on_page' has wrong type; need 'int' but got "
            + f"'{type(resume_on_page).__name__}'"
        )
    if resume_on_page != 1:
        slk_call = slk_call + " --resume-on-page " + str(resume_on_page)

    # --results-per-page
    if not isinstance(results_per_page, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'results_per_page' has wrong type; need 'int' but got "
            + f"'{type(results_per_page).__name__}'"
        )
    if results_per_page not in [-1, 0, 1000]:
        slk_call = slk_call + " --results-per-page " + str(results_per_page)

    # --end-on-page
    if not isinstance(end_on_page, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'end_on_page' has wrong type; need 'int' but got "
            + f"'{type(end_on_page).__name__}'"
        )
    if end_on_page != -1:
        slk_call = slk_call + " --end-on-page " + str(end_on_page)

    # --save-mode
    if not isinstance(save_mode, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'save_mode' has wrong type; need 'bool' but got "
            + f"'{type(save_mode).__name__}'"
        )
    if save_mode:
        slk_call = slk_call + " --save-mode"

    # --verbose
    if not isinstance(verbose, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'verbose' has wrong type; need 'bool' but got "
            + f"'{type(verbose).__name__}'"
        )
    if verbose:
        slk_call = slk_call + " --verbose"

    # --json
    if not isinstance(json, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json' has wrong type; need 'bool' but got "
            + f"'{type(json).__name__}'"
        )
    if json:
        slk_call = slk_call + " --json"

    # --recursive
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'recursive' has wrong type; need 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    if recursive:
        slk_call = slk_call + " --recursive"

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=0,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=0,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=0,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


# TODO: nice wrapper
def result_verify_job_raw(
    job_id: Union[int, str],
    header: bool = False,
    json: bool = False,
    json_no_source: bool = False,
    number_errors: bool = False,
    number_sources: bool = False,
    sources: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """prints result verification job information

    :param job_id: job id
    :type job_id: ``int`` or ``str``
    :param header: print header
    :type header: ``bool``
    :param json: print json
    :type json: ``bool``
    :param json_no_source: print json without source
    :type json_no_source: ``bool``
    :param number_errors: print number of errors
    :type number_errors: ``bool``
    :param number_sources: print number of sources
    :type number_sources: ``bool``
    :param sources: print sources
    :type sources: ``bool``
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "result_verify_job"]

    # job_id
    if isinstance(job_id, int):
        slk_call.append(str(job_id))
    elif isinstance(job_id, str):
        try:
            slk_call.append(str(int(job_id)))
        except ValueError:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'job_id' cannot be processed; need 'str', which holds an integer "
                + f"number, or 'int' but got '{type(job_id).__name__}' with value '{job_id}'"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'job_id' has wrong type; need 'str' or 'int' but got "
            + f"'{type(job_id).__name__}'"
        )

    # --header
    if not isinstance(header, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'header' has wrong type; need 'bool' but got "
            + f"'{type(header).__name__}'"
        )
    if header:
        slk_call.append("--header")

    # --json
    if not isinstance(json, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json' has wrong type; need 'bool' but got "
            + f"'{type(json).__name__}'"
        )
    if json:
        slk_call.append("--json")

    # --json-no-source
    if not isinstance(json_no_source, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json_no_source' has wrong type; need 'bool' but got "
            + f"'{type(json_no_source).__name__}'"
        )
    if json_no_source:
        slk_call.append("--json-no-source")

    # --number-errors
    if not isinstance(number_errors, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'number_errors' has wrong type; need 'bool' but got "
            + f"'{type(number_errors).__name__}'"
        )
    if number_errors:
        slk_call.append("--number-errors")

    # --number-sources
    if not isinstance(number_sources, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'number_sources' has wrong type; need 'bool' but got "
            + f"'{type(number_sources).__name__}'"
        )
    if number_sources:
        slk_call.append("--number-sources")

    # --sources
    if not isinstance(sources, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'sources' has wrong type; need 'bool' but got "
            + f"'{type(sources).__name__}'"
        )
    if sources:
        slk_call.append("--sources")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import inspect
import subprocess
import typing
from pathlib import Path
from typing import Union

from ..base import get_resource_path, searchid_exists
from ..pyslk_exceptions import PySlkException
from ..raw import exists_raw, resource_path_raw, resource_type_raw

__all__ = [
    "get_resource_type",
    "is_file",
    "is_namespace",
    "resource_exists",
    "_check_resource",
    "_check_resource_and_search_id",
]


def resource_exists(resource: Union[str, Path, int]) -> bool:
    """Check if resource exists and return True/False

    :param resource: namespace or resource
    :type resource: ``str`` or ``path-like``
    :returns: True if file exists; False otherwise
    :rtype: ``bool``
    """
    # determine whether we have resource id or resource path
    resource_id: Union[int, None] = None
    resource_path: Union[Path, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' has wrong type; need 'str', 'int' or 'Path' but "
            + f"got {type(resource).__name__}"
        )
    # run slk_helpers command
    if resource_path is not None:
        output: subprocess.CompletedProcess = exists_raw(
            str(resource_path), return_type=2
        )
    else:
        # resource_id is not None
        output: subprocess.CompletedProcess = resource_path_raw(
            resource_id, return_type=2
        )
    # evaluate output
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def _check_resource_and_search_id(
    calling_function: str,
    resource: Union[str, int, Path, list, None] = None,
    search_id: Union[str, int, None] = None,
) -> typing.Optional[tuple[Union[Path, list[Path]], int, int]]:
    """Checks whether the input arguments 'resource' and 'search_id' are properly set

    Thrown errors

    * ``PySlkException``
    * ``TypeError``
    * ``ValueError``

    :param calling_function: name of the functions which calls this function
    :type calling_function: ``str``
    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: True if one or more files are flagged; False otherwise; if 'details': dictionary with two keys
        'flag_partial' and 'no_flag_partial' which each have a list of files as value
    :rtype: ``bool`` or ``dict`` or ``None``
    """
    # check if one of resource and search_id is set
    if resource is None and search_id is None:
        raise ValueError(
            f"pyslk.{calling_function}: either 'resource' xor 'search_id' have to be non-None values"
        )
    if resource is not None and search_id is not None:
        raise ValueError(
            f"pyslk.{calling_function}: either 'resource' xor 'search_id' have to be non-None values, not both"
        )
    # ~~~~~~~~~ check resource ~~~~~~~~~
    if resource is not None:
        # basic type checking
        if not isinstance(resource, (str, int, Path, list)):
            raise TypeError(
                f"pyslk.{calling_function}: 'resource' has wrong type; need 'str', 'int', 'Path' or 'list' but "
                + f"got {type(resource).__name__}"
            )
        if isinstance(resource, list):
            if len(resource) == 0:
                # check if resource exists and if empty list
                return None
            elif isinstance(resource, list) and len(resource) == 1:
                # make list to non-list if list has length of 1
                resource = resource[0]
            else:
                # here, we only have lists with length > 1
                if any([isinstance(i, int) for i in resource]):
                    # check if we have multiple resources and at least one is an int
                    raise TypeError(
                        f"pyslk.{calling_function}: 'resource' is a list with more than one item; at least one item "
                        + "is of type 'int' (== resource id); however, only one resource id is allowed"
                    )
                if any([not isinstance(i, (str, Path)) for i in resource]):
                    raise TypeError(
                        f"pyslk.{calling_function}: 'resource' is a list; the elements in the list have to be 'str' or "
                        + f"path-like but got: {', '.join([type(res).__name__ for res in resource])}"
                    )
        elif not resource_exists(resource):
            # we arrive here if we are not a list;
            # check if resource exists
            return None
    # ~~~~~~~~~ check search ~~~~~~~~~
    # check if search_id exists if non-None
    if search_id is not None and not searchid_exists(search_id):
        return None
    # ~~~~~~~~~ convert values ~~~~~~~~~
    resource_id: Union[int, None] = None
    resource_path: Union[Path, list, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, list):
        # note: we checked already above whether all elements of the list are 'str' or path-like
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{calling_function}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    if isinstance(search_id, str):
        # resource can either be a resource id or a resource path
        try:
            search_id = int(search_id)
        except ValueError:
            raise TypeError(
                f"pyslk.{calling_function}: cannot convert 'search_id' from 'str' to 'int'"
            )
    # return proper values
    return resource_path, resource_id, search_id


def _check_resource(
    calling_function: str,
    resource: Union[str, int, Path, list],
) -> typing.Optional[tuple[Union[Path, list[Path]], int]]:
    """Checks whether the input argument 'resource' is properly set

    Thrown errors

    * ``PySlkException``
    * ``TypeError``
    * ``ValueError``

    :param calling_function: name of the functions which calls this function
    :type calling_function: ``str``
    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: True if one or more files are flagged; False otherwise; if 'details': dictionary with two keys
        'flag_partial' and 'no_flag_partial' which each have a list of files as value
    :rtype: ``bool`` or ``dict`` or ``None``
    """
    # ~~~~~~~~~ check resource ~~~~~~~~~
    if resource is not None:
        # basic type checking
        if not isinstance(resource, (str, int, Path, list)):
            raise TypeError(
                f"pyslk.{calling_function}: 'resource' has wrong type; need 'str', 'int', 'Path' or 'list' but "
                + f"got {type(resource).__name__}"
            )
        if isinstance(resource, list):
            if len(resource) == 0:
                # check if resource exists and if empty list
                return None
            elif isinstance(resource, list) and len(resource) == 1:
                # make list to non-list if list has length of 1
                resource = resource[0]
            else:
                # here, we only have lists with length > 1
                if any([isinstance(i, int) for i in resource]):
                    # check if we have multiple resources and at least one is an int
                    raise TypeError(
                        f"pyslk.{calling_function}: 'resource' is a list with more than one item; at least one item "
                        + "is of type 'int' (== resource id); however, only one resource id is allowed"
                    )
                if any([not isinstance(i, (str, Path)) for i in resource]):
                    raise TypeError(
                        f"pyslk.{calling_function}: 'resource' is a list; the elements in the list have to be 'str' or "
                        + f"path-like but got: {', '.join([type(res).__name__ for res in resource])}"
                    )
        elif not resource_exists(resource):
            # we arrive here if we are not a list;
            # check if resource exists
            return None
    # ~~~~~~~~~ convert values ~~~~~~~~~
    resource_id: Union[int, None] = None
    resource_path: Union[Path, list, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, list):
        # note: we checked already above whether all elements of the list are 'str' or path-like
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{calling_function}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    # return proper values
    return resource_path, resource_id


def get_resource_type(resource: Union[str, int, Path]) -> typing.Optional[str]:
    """Get type of resource

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: type of the resource; None if resource does not exist
    :rtype: ``str`` or ``None``
    """
    if not resource_exists(resource):
        return None
    resource_id: Union[int, None] = None
    resource_path: Union[Path, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' has wrong type; need 'str', 'int' or 'Path' but "
            + f"got {type(resource).__name__}"
        )
    if resource_path is not None and not resource_exists(resource_path):
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' was considered as resource path and does not exist: {resource}"
        )
    if resource_id is not None and not get_resource_path(resource_id):
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' was considered as resource id and does not exist: {resource}"
        )
    output: subprocess.CompletedProcess = resource_type_raw(
        resource_path, resource_id, return_type=2
    )
    if output.returncode == 0:
        return output.stdout.decode("utf-8").rstrip()
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_namespace(resource: Union[str, int, Path, None] = None) -> typing.Optional[bool]:
    """Returns True if resource is a namespace

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: True if resource is a namespace; else False
    :rtype: ``str`` or ``None``
    """
    # run get_resource_type
    output = get_resource_type(resource)
    if output is None:
        return None
    # check the type
    if output.lower() == "namespace":
        return True
    return False


def is_file(resource: Union[str, int, Path, None] = None) -> typing.Optional[bool]:
    """Returns True if resource is a file

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: True if resource is a file; else False
    :rtype: ``str`` or ``None``
    """
    # run get_resource_type
    output = get_resource_type(resource)
    if output is None:
        return None
    # check the type
    if output.lower() == "file":
        return True
    return False



# flake8: noqa
"""
The pyslk package provides Python wrapper functions for the slk and the
slk_helpers. The slk is the official command line interface (cli) of the
StrongLink system by StrongBox Data Solutions. The slk_helpers was developed
at DKRZ and provides some additional functionality beyond the slk.

* **pyslk.pyslk:** basic wrappers for nearly all slk and slk_helpers commands
* **pyslk.parsers:** extended wrappers for a few slk and slk_helpers commands
* **pyslk.utils:** some utility functions
* **pyslk.exceptions:** one exception class used in this package
"""
import importlib.metadata

from .base import (
    FileGroup,
    GroupCollection,
    StatusJob,
    cli_versions,
    expiration_date,
    get_resource_path,
    get_search_status,
    hostname,
    is_search_incomplete,
    is_search_successful,
    list_clone_file,
    list_clone_search,
)
from .base import ls
from .base import ls as list
from .base import (
    search,
    search_immediately,
    searchid_exists,
    session,
    total_number_search_results,
    valid_session,
    valid_token,
    version_slk,
    version_slk_helpers,
)
from .config import get, module_load, set
from .constants import (
    MAX_RETRIES_SUBMIT_JOBS_BASIC,
    MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE,
    MIN_VERSION_SLK,
    MIN_VERSION_SLK_HELPERS,
    PYSLK_DEFAULT_LIST_CLONE_COLUMNS,
    PYSLK_DEFAULT_LIST_COLUMN_TYPES,
    PYSLK_DEFAULT_LIST_COLUMNS,
    PYSLK_DEFAULT_LIST_COLUMNS_3_3_21,
    PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
    PYSLK_FILE_SIZE_UNITS,
    PYSLK_LOGGER,
    PYSLK_SEARCH_DELAY,
    PYSLK_WILDCARDS,
    SEARCH_QUERY_OPERATORS,
    SLK,
    SLK_HELPERS,
    SLK_SYSTEM_CONFIG,
    SLK_USER_CONFIG,
    SLK_USER_LOG,
)
from .core import (
    access_hsm,
    arch_size,
    chgrp,
    chmod,
    chown,
    count_tapes,
    count_tapes_with_multi_tape_files,
    count_tapes_with_single_tape_files,
    delete,
    gen_file_query,
    gen_file_query_as_dict,
    gen_search_query,
    get_checksum,
    get_metadata,
    get_rcrs,
    get_resource_id,
    get_resource_mtime,
    get_resource_permissions,
    get_resource_size,
    get_resource_tape,
    get_resource_type,
    get_storage_information,
    get_tag,
    get_tape_barcode,
    get_tape_id,
    get_tape_status,
    group_files_by_tape,
    has_no_flag_partial,
    has_no_flag_partial_details,
    hsm2json,
    hsm2json_dict,
    hsm2json_file,
    is_cached,
    is_cached_details,
    is_file,
    is_namespace,
    is_on_tape,
    is_on_tape_details,
    is_tape_available,
    json2hsm,
    json_dict2hsm,
    json_file2hsm,
    json_str2hsm,
    makedirs,
    mkdir,
    move,
    rename,
    resource_exists,
    set_tag,
    tape_exists,
    tape_status,
)
from .jobs import (
    SubmittedJobs,
    get_bad_files_verify_job,
    get_checked_resources_verify_job,
    get_job_status,
    get_result_verify_job,
    is_job_finished,
    is_job_processing,
    is_job_queued,
    is_job_successful,
    job_exists,
    job_queue,
    submit_verify_job,
)
from .pyslk_exceptions import (
    ArchiveError,
    HostNotReachableError,
    PySlkBadFileError,
    PySlkBadProcessError,
    PySlkEmptyInputError,
    PySlkException,
    PySlkNoValidLoginTokenError,
    SizeMismatchError,
    SlkIOError,
    check_for_errors,
)
from .pyslk_warnings import SlkWarning
from .raw import (
    archive_raw,
    checksum_raw,
    chmod_raw,
    delete_raw,
    exists_raw,
    gen_file_query_raw,
    gen_search_query_raw,
    group_files_by_tape_raw,
    group_raw,
    has_no_flag_partial_raw,
    hostname_raw,
    hsm2json_raw,
    is_on_tape_raw,
    iscached_raw,
    job_exists_raw,
    job_queue_raw,
    job_report_raw,
    job_status_raw,
    json2hsm_raw,
    list_clone_file_raw,
    list_clone_search_raw,
    list_raw,
    metadata_raw,
    mkdir_raw,
    move_raw,
    owner_raw,
    print_rcrs_raw,
    recall2_raw,
    recall_raw,
    rename_raw,
    resource_mtime_raw,
    resource_path_raw,
    resource_permissions_raw,
    resource_tape_raw,
    resource_type_raw,
    result_verify_job_raw,
    retrieve2_raw,
    retrieve_raw,
    search_immediately_raw,
    search_incomplete_raw,
    search_raw,
    search_status_raw,
    search_successful_raw,
    searchid_exists_raw,
    session_raw,
    size_raw,
    submit_verify_job_raw,
    tag_raw,
    tape_barcode_raw,
    tape_exists_raw,
    tape_id_raw,
    tape_library_raw,
    tape_status_raw,
    tnsr_raw,
    total_number_search_results_raw,
    version_slk_helpers_raw,
    version_slk_raw,
)
from .retry import log_attempt_number, try_archive
from .transfer import archive, recall_dev, recall_single, retrieve, retrieve_improved
from .utils import (
    FakeProc,
    access_local,
    construct_dst_from_src,
    convert_expiration_date,
    get_recall_job_id,
    get_slk_pid,
    is_search_id,
    login_valid,
    run_slk,
    which,
)
from .version import slk_helpers_version, slk_version

__all__ = [
    "ArchiveError",
    "FakeProc",
    "FileGroup",
    "GroupCollection",
    "HostNotReachableError",
    "MAX_RETRIES_SUBMIT_JOBS_BASIC",
    "MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE",
    "MIN_VERSION_SLK",
    "MIN_VERSION_SLK_HELPERS",
    "PYSLK_DEFAULT_LIST_CLONE_COLUMNS",
    "PYSLK_DEFAULT_LIST_COLUMNS",
    "PYSLK_DEFAULT_LIST_COLUMNS_3_3_21",
    "PYSLK_DEFAULT_LIST_COLUMN_TYPES",
    "PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS",
    "PYSLK_FILE_SIZE_UNITS",
    "PYSLK_LOGGER",
    "PYSLK_SEARCH_DELAY",
    "PYSLK_WILDCARDS",
    "PySlkBadFileError",
    "PySlkBadProcessError",
    "PySlkEmptyInputError",
    "PySlkException",
    "PySlkNoValidLoginTokenError",
    "SEARCH_QUERY_OPERATORS",
    "SLK",
    "SLK_HELPERS",
    "SLK_SYSTEM_CONFIG",
    "SLK_USER_CONFIG",
    "SLK_USER_LOG",
    "SizeMismatchError",
    "SlkIOError",
    "SlkWarning",
    "StatusJob",
    "SubmittedJobs",
    "access_hsm",
    "access_local",
    "arch_size",
    "archive",
    "archive_raw",
    "check_for_errors",
    "checksum_raw",
    "chgrp",
    "chmod",
    "chmod_raw",
    "chown",
    "cli_versions",
    "construct_dst_from_src",
    "convert_expiration_date",
    "count_tapes",
    "count_tapes_with_multi_tape_files",
    "count_tapes_with_single_tape_files",
    "delete",
    "delete_raw",
    "exists_raw",
    "expiration_date",
    "gen_file_query",
    "gen_file_query_as_dict",
    "gen_file_query_raw",
    "gen_search_query",
    "gen_search_query_raw",
    "get",
    "get_bad_files_verify_job",
    "get_checked_resources_verify_job",
    "get_checksum",
    "get_job_status",
    "get_metadata",
    "get_rcrs",
    "get_recall_job_id",
    "get_resource_id",
    "get_resource_mtime",
    "get_resource_path",
    "get_resource_permissions",
    "get_resource_size",
    "get_resource_tape",
    "get_resource_type",
    "get_result_verify_job",
    "get_search_status",
    "get_slk_pid",
    "get_storage_information",
    "get_tag",
    "get_tape_barcode",
    "get_tape_id",
    "get_tape_status",
    "group_files_by_tape",
    "group_files_by_tape_raw",
    "group_raw",
    "has_no_flag_partial",
    "has_no_flag_partial_details",
    "has_no_flag_partial_raw",
    "hostname",
    "hostname_raw",
    "hsm2json",
    "hsm2json_dict",
    "hsm2json_file",
    "hsm2json_raw",
    "is_cached",
    "is_cached_details",
    "is_file",
    "is_job_finished",
    "is_job_processing",
    "is_job_queued",
    "is_job_successful",
    "is_namespace",
    "is_on_tape",
    "is_on_tape_details",
    "is_on_tape_raw",
    "is_search_id",
    "is_search_incomplete",
    "is_search_successful",
    "is_tape_available",
    "iscached_raw",
    "job_exists",
    "job_exists_raw",
    "job_queue",
    "job_queue_raw",
    "job_report_raw",
    "job_status_raw",
    "json2hsm",
    "json2hsm_raw",
    "json_dict2hsm",
    "json_file2hsm",
    "json_str2hsm",
    "list",
    "list_clone_file",
    "list_clone_file_raw",
    "list_clone_search",
    "list_clone_search_raw",
    "list_raw",
    "log_attempt_number",
    "login_valid",
    "ls",
    "makedirs",
    "metadata_raw",
    "mkdir",
    "mkdir_raw",
    "module_load",
    "move",
    "move_raw",
    "owner_raw",
    "print_rcrs_raw",
    "recall_dev",
    "recall_raw",
    "recall_single",
    "recall2_raw",
    "rename",
    "rename_raw",
    "resource_exists",
    "resource_mtime_raw",
    "resource_path_raw",
    "resource_permissions_raw",
    "resource_tape_raw",
    "resource_type_raw",
    "result_verify_job_raw",
    "retrieve",
    "retrieve_improved",
    "retrieve_raw",
    "retrieve2_raw",
    "run_slk",
    "search",
    "search_immediately",
    "search_immediately_raw",
    "search_incomplete_raw",
    "search_raw",
    "search_status_raw",
    "search_successful_raw",
    "searchid_exists",
    "searchid_exists_raw",
    "session",
    "session_raw",
    "set",
    "set_tag",
    "size_raw",
    "slk_helpers_version",
    "slk_version",
    "submit_verify_job",
    "submit_verify_job_raw",
    "tag_raw",
    "tape_barcode_raw",
    "tape_exists",
    "tape_exists_raw",
    "tape_id_raw",
    "tape_library_raw",
    "tape_status",
    "tape_status_raw",
    "tnsr_raw",
    "total_number_search_results",
    "total_number_search_results_raw",
    "try_archive",
    "valid_session",
    "valid_token",
    "version_slk",
    "version_slk_helpers",
    "version_slk_helpers_raw",
    "version_slk_raw",
    "which",
]

__version__ = importlib.metadata.version("pyslk")
ENV = config.module_load()



import inspect
import json
import subprocess
import typing
import warnings
from pathlib import Path
from typing import Union

from ..pyslk_exceptions import PySlkException
from ..raw import (
    is_on_tape_raw,
    iscached_raw,
    print_rcrs_raw,
    tape_barcode_raw,
    tape_exists_raw,
    tape_id_raw,
    tape_status_raw,
)
from .resource_extras import _check_resource, _check_resource_and_search_id

__all__ = [
    "_cached",
    "is_cached",
    "is_cached_details",
    "_on_tape",
    "is_on_tape",
    "is_on_tape_details",
    "get_tape_barcode",
    "get_tape_status",
    "tape_exists",
    "get_tape_id",
    "get_storage_information",
    "get_rcrs",
    "tape_status",
    "is_tape_available",
]


def _cached(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
    details: bool = False,
) -> typing.Optional[Union[bool, dict[str, list[Path]]]]:
    """Check if whether file(s) is/are in HSM cache or not; returns True/False or a dict with keys 'cached' and
        'not_cached', depending on 'details'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :param details: print caching status of each file; if not set, output is False if not file in cache and else True
    :type details: ``bool``
    :returns: True if file is in cache; False otherwise; if 'details': dictionary with an entry per file
    :rtype: ``bool`` or ``dict`` or ``None``
    """
    # check a parse the input values
    check_result = _check_resource_and_search_id(
        inspect.stack()[0][3], resource, search_id
    )
    # if we received None=> return None
    if check_result is None:
        return None
    # extract values from resource path, resource id and search id
    resource_path, resource_id, search_id = check_result
    # run raw function
    output: subprocess.CompletedProcess = iscached_raw(
        resource_path,
        resource_id,
        search_id,
        recursive=True,
        verbose=False,
        double_verbose=details,
        return_type=2,
    )
    if output.returncode in [0, 1]:
        if not details:
            if output.returncode == 0:
                return True
            elif output.returncode == 1:
                return False
        else:
            # [:-1] because we drop the last row
            return {
                "cached": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "not" not in o and "unclear" not in o
                ],
                "not_cached": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "not" in o
                ],
                "unclear": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "unclear" in o
                ],
            }
    else:
        # output.returncode is neither 0 nor 1
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_cached(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
) -> bool:
    """Check if whether file(s) is/are in HSM cache or not; returns True/False

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: True if all files are in cache; False otherwise
    :rtype: ``bool``
    """
    return _cached(resource, search_id, details=False)


def is_cached_details(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
) -> dict[str, list[Path]]:
    """Check if whether file(s) is/are in HSM cache or not; returns dict with keys 'cached' and 'not_cached'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: dictionary with two keys 'cached' and 'not_cached' which each have a list of files as value
    :rtype: ``dict[str, list[Path]]``
    """
    return _cached(resource, search_id, details=True)


def _on_tape(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
    details: bool = False,
) -> typing.Optional[Union[bool, dict[str, list[Path]]]]:
    """Check if whether file(s) is/are stored on tape or not; returns True/False or a dict with keys 'on_tape' and
        'not_on_tape', depending on 'details'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :param details: print tape storage status of each file; if not set, output is False if not file on tape
    :type details: ``bool``
    :returns: True if file is on tape; False otherwise; if 'details': dictionary with an entry per file
    :rtype: ``bool`` or ``dict`` or ``None``
    """
    # check a parse the input values
    check_result = _check_resource_and_search_id(
        inspect.stack()[0][3], resource, search_id
    )
    # if we received None=> return None
    if check_result is None:
        return None
    # extract values from resource path, resource id and search id
    resource_path, resource_id, search_id = check_result
    # run raw function
    output: subprocess.CompletedProcess = is_on_tape_raw(
        resource_path,
        resource_id,
        search_id,
        recursive=True,
        verbose=False,
        double_verbose=details,
        return_type=2,
    )
    if output.returncode in [0, 1]:
        if not details:
            if output.returncode == 0:
                return True
            elif output.returncode == 1:
                return False
        else:
            # [:-1] because we drop the last row
            return {
                "on_tape": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "not" not in o
                ],
                "not_on_tape": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "not" in o
                ],
            }
    else:
        # output.returncode is neither 0 nor 1
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_on_tape(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
) -> bool:
    """Check if whether file(s) is/are stored on tape; returns True/False

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: True if all files are stored on tape; False otherwise
    :rtype: ``bool``
    """
    return _on_tape(resource, search_id, details=False)


def is_on_tape_details(
    resource: Union[str, int, Path, None] = None,
    search_id: Union[str, int, None] = None,
) -> dict[str, list[Path]]:
    """Check if whether file(s) is/are stored on tape or not; returns dict with keys 'on_tape' and 'not_on_tape'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: dictionary with two keys 'on_tape' and 'not_on_tape' which each have a list of files as value
    :rtype: ``dict[str, list[Path]]``
    """
    return _on_tape(resource, search_id, details=True)


def get_tape_barcode(tape_id: Union[int, str]) -> typing.Optional[str]:
    """return tape barcode for provided tape id

    :param tape_id: id of a tape in the tape library
    :type tape_id: ``int`` or ``str``
    :returns: True if tape exists; False otherwise
    :rtype: ``bool``
    """
    if not isinstance(tape_id, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'int'; got '{type(tape_id).__name__}'"
        )
    output: subprocess.CompletedProcess = tape_barcode_raw(tape_id, return_type=2)
    if output.returncode == 0:
        return output.stdout.decode("utf-8").rstrip()
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def tape_exists(tape: Union[int, str]) -> bool:
    """Check if tape exists

    :param tape: id or barcode of a tape in the tape library
    :type tape: ``int`` or ``str``
    :returns: True if tape exists; False otherwise
    :rtype: ``bool``
    """
    if not isinstance(tape, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'str' or 'int'; got '{type(tape).__name__}'"
        )
    tape_id: Union[int, None] = None
    tape_barcode: Union[str, None] = None
    if isinstance(tape, int):
        tape_id = tape
    elif isinstance(tape, str):
        # tape can either be a tape id or a tape barcode
        try:
            tape_id = int(tape)
        except ValueError:
            tape_barcode = tape
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'tape' has wrong type; need 'str' or 'int' but "
            + f"got {type(tape).__name__}"
        )

    output: subprocess.CompletedProcess = tape_exists_raw(
        tape_id, tape_barcode, return_type=2
    )
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_library(tape: Union[int, str]) -> str:
    """Check in which library a tape is located

    :param tape: id or barcode of a tape in the tape library
    :type tape: ``int`` or ``str``
    :returns: name of the library in which the tape is located
    :rtype: ``str``
    """
    if not isinstance(tape, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'str' or 'int'; got '{type(tape).__name__}'"
        )
    tape_id: Union[int, None] = None
    tape_barcode: Union[str, None] = None
    input_type: str
    if isinstance(tape, int):
        tape_id = tape
        input_type = "tape id"
    elif isinstance(tape, str):
        # tape can either be a tape id or a tape barcode
        try:
            tape_id = int(tape)
            input_type = "tape id"
        except ValueError:
            tape_barcode = tape
            input_type = "tape barcode"
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'tape' has wrong type; need 'str' or 'int' but "
            + f"got {type(tape).__name__}"
        )

    output: subprocess.CompletedProcess = tape_exists_raw(
        tape_id, tape_barcode, return_type=2
    )
    if output.returncode == 0:
        return output.stdout.decode("utf-8").rstrip()
    elif output.returncode == 1:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: tape not found: {tape} (interpreted as {input_type})"
        )
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_tape_id(tape_barcode: str) -> typing.Optional[int]:
    """return tape id for provided tape barcode

    :param tape_barcode: barcode of a tape in the tape library
    :type tape_barcode: ``str``
    :returns: tape id; None otherwise
    :rtype: ``int``
    """
    if not isinstance(tape_barcode, str):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'str'; got '{type(tape_barcode).__name__}'"
        )
    output: subprocess.CompletedProcess = tape_id_raw(tape_barcode, return_type=2)
    if output.returncode == 0:
        return int(output.stdout.decode("utf-8").rstrip())
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def tape_status(tape: Union[int, str], details: bool = False) -> typing.Optional[str]:
    """Check the status of a tape

    :param tape: id  or barcode of a tape in the tape library
    :type tape: ``int`` or ``str``
    :param details: print a more detailed description of the retrieval status
    :type details: ``bool``
    :returns: status of the tape; None if tape does not exist
    :rtype: ``str`` or ``None``
    """
    return get_tape_status(tape, details)


def get_tape_status(
    tape: Union[int, str], details: bool = False
) -> typing.Optional[str]:
    """Check the status of a tape

    :param tape: id  or barcode of a tape in the tape library
    :type tape: ``int`` or ``str``
    :param details: print a more detailed description of the retrieval status
    :type details: ``bool``
    :returns: status of the tape; None if tape does not exist
    :rtype: ``str`` or ``None``
    """
    if not tape_exists(tape):
        return None
    tape_id: Union[int, None] = None
    tape_barcode: Union[str, None] = None
    if isinstance(tape, int):
        tape_id = tape
    elif isinstance(tape, str):
        # tape can either be a tape id or a tape barcode
        try:
            tape_id = int(tape)
        except ValueError:
            tape_barcode = tape
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'tape' has wrong type; need 'str' or 'int' but "
            + f"got {type(tape).__name__}"
        )

    output: subprocess.CompletedProcess = tape_status_raw(
        tape_id, tape_barcode, details, return_type=2
    )
    if output.returncode in [0, 1]:
        if not details:
            return output.stdout.decode("utf-8").rstrip()
        else:
            return output.stdout.decode("utf-8").split("\n")[1]
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_tape_available(tape: int) -> typing.Optional[bool]:
    """Check if tape is available

    :param tape: id  or barcode of a tape in the tape library
    :type tape: ``int`` or ``str``
    :returns: True if tape is available for recalls/retrievals; else False; None if tape does not exist
    :rtype: ``bool`` or ``None``
    """
    if not tape_exists(tape):
        return None
    tape_id: Union[int, None] = None
    tape_barcode: Union[str, None] = None
    if isinstance(tape, int):
        tape_id = tape
    elif isinstance(tape, str):
        # tape can either be a tape id or a tape barcode
        try:
            tape_id = int(tape)
        except ValueError:
            tape_barcode = tape
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'tape' has wrong type; need 'str' or 'int' but "
            + f"got {type(tape).__name__}"
        )
    output: subprocess.CompletedProcess = tape_status_raw(
        tape_id, tape_barcode, details=False, return_type=2
    )
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_storage_information(resource: Union[str, int, Path]) -> typing.Optional[dict]:
    """prints resource content record (rcr) information

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: True if tape exists; False otherwise
    :rtype: ``bool``
    """
    return get_rcrs(resource)


def get_rcrs(resource: Union[str, int, Path]) -> typing.Optional[dict]:
    """prints resource content record (rcr) information

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :returns: storage information when exists; None otherwise
    :rtype: ``dict``
    """
    # check a parse the input values
    check_result = _check_resource(inspect.stack()[0][3], resource)
    # if we received None=> return None
    if check_result is None:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: parameter 'resource' not properly specified: {resource}"
        )
    # extract values from resource path, resource id and search id
    resource_path, resource_id = check_result
    output: subprocess.CompletedProcess = print_rcrs_raw(
        resource_path, resource_id, json=True, return_type=2
    )
    if output.returncode == 0:
        return json.loads(output.stdout.decode("utf-8").rstrip())
    elif output.returncode == 1:
        warnings.warn(
            f"pyslk.{inspect.stack()[0][3]}: storage information incomplete; returning what is available"
        )
        return json.loads(output.stdout.decode("utf-8").rstrip())
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
